import logging

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.services.remote_server import MessageAux


class ConditionInducer:
    IDENTIFIER = 'com.apple.instruments.server.services.ConditionInducer'

    def __init__(self, dvt):
        self.logger = logging.getLogger(__name__)
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def list(self) -> list:
        self._channel.availableConditionInducers()
        return self._channel.receive_plist()

    def set(self, profile_identifier):
        for group in self.list():
            for profile in group.get('profiles'):
                if profile_identifier == profile.get('identifier'):
                    self.logger.info(profile.get('description'))
                    self._channel.enableConditionWithIdentifier_profileIdentifier_(
                        MessageAux().append_obj(group.get('identifier')).append_obj(profile.get('identifier')))
                    return
        raise PyMobileDevice3Exception('Invalid profile identifier')

    def clear(self):
        self._channel.disableActiveCondition()



import logging

from cached_property import cached_property
from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.restore.img4 import stitch_component
from pymobiledevice3.restore.tss import TSSResponse


class Component:
    def __init__(self, build_identity, name: str, tss: TSSResponse = None, data: bytes = None, path: str = None):
        self.logger = logging.getLogger(__name__)
        self._tss = tss
        self.build_identity = build_identity
        self.name = name
        self._data = data
        self._path = path

    @cached_property
    def path(self):
        if self._path:
            return self._path

        path = None
        if self._tss:
            path = self._tss.get_path_by_entry(self.name)

            if path is None:
                self.logger.debug(f'NOTE: No path for component {self.name} in TSS, will fetch from build_identity')

        if path is None:
            path = self.build_identity.get_component_path(self.name)

        if path is None:
            raise PyMobileDevice3Exception(f'Failed to find component path for: {self.name}')

        return path

    @cached_property
    def data(self):
        if self._data is None:
            return self.build_identity.build_manifest.ipsw.read(self.path)
        return self._data

    @cached_property
    def personalized_data(self):
        if self._tss is None:
            raise PyMobileDevice3Exception(f'TSS ticket must be supplied for personalizing component: {self.name}')

        # stitch ApImg4Ticket into IMG4 file
        return stitch_component(self.name, self.data, self._tss.ap_img4_ticket)



import asyncio
from dataclasses import dataclass, fields
from enum import Enum
import logging
import uuid
import json

import nest_asyncio

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol
from pymobiledevice3.services.web_protocol.automation_session import AutomationSession
from pymobiledevice3.exceptions import WebInspectorNotEnabled, RemoteAutomationNotEnabled

SAFARI = 'com.apple.mobilesafari'


def key_to_pid(key):
    return int(key.split(':')[1])


class WirTypes(Enum):
    AUTOMATION = 'WIRTypeAutomation'
    ITML = 'WIRTypeITML'
    JAVASCRIPT = 'WIRTypeJavaScript'
    PAGE = 'WIRTypePage'
    SERVICE_WORKER = 'WIRTypeServiceWorker'
    WEB = 'WIRTypeWeb'
    WEB_PAGE = 'WIRTypeWebPage'
    AUTOMATICALLY_PAUSE = 'WIRAutomaticallyPause'


class AutomationAvailability(Enum):
    NOT_AVAILABLE = 'WIRAutomationAvailabilityNotAvailable'
    AVAILABLE = 'WIRAutomationAvailabilityAvailable'
    UNKNOWN = 'WIRAutomationAvailabilityUnknown'


@dataclass
class Page:
    id_: int
    type_: WirTypes
    web_url: str = ''
    web_title: str = ''
    automation_is_paired_key: bool = False
    automation_name: str = ''
    automation_version: str = ''
    automation_session_id: str = ''
    automation_connection_id: str = ''

    @classmethod
    def from_page_dictionary(cls, page_dict):
        p = cls(page_dict['WIRPageIdentifierKey'], WirTypes(page_dict['WIRTypeKey']))
        if p.type_ in (WirTypes.WEB, WirTypes.WEB_PAGE):
            p.web_title = page_dict['WIRTitleKey']
            p.web_url = page_dict['WIRURLKey']
        if p.type_ == WirTypes.AUTOMATION:
            p.automation_is_paired_key = page_dict['WIRAutomationTargetIsPairedKey']
            p.automation_name = page_dict['WIRAutomationTargetNameKey']
            p.automation_version = page_dict['WIRAutomationTargetVersionKey']
            p.automation_session_id = page_dict['WIRSessionIdentifierKey']
            if 'WIRConnectionIdentifierKey' in page_dict:
                p.automation_connection_id = page_dict['WIRConnectionIdentifierKey']
        return p

    def update(self, page_dict):
        new_p = self.from_page_dictionary(page_dict)
        for field in fields(self):
            setattr(self, field.name, getattr(new_p, field.name))


@dataclass
class Application:
    id_: str
    bundle: str
    pid: int
    name: str
    availability: AutomationAvailability
    active: int
    proxy: bool
    ready: bool
    host: str = ''

    @classmethod
    def from_application_dictionary(cls, app_dict):
        return cls(
            app_dict['WIRApplicationIdentifierKey'],
            app_dict['WIRApplicationBundleIdentifierKey'],
            key_to_pid(app_dict['WIRApplicationIdentifierKey']),
            app_dict['WIRApplicationNameKey'],
            AutomationAvailability(app_dict['WIRAutomationAvailabilityKey']),
            app_dict['WIRIsApplicationActiveKey'],
            app_dict['WIRIsApplicationProxyKey'],
            app_dict['WIRIsApplicationReadyKey'],
            app_dict.get('WIRHostApplicationIdentifierKey', ''),
        )


class WebinspectorService:
    SERVICE_NAME = 'com.apple.webinspector'

    def __init__(self, lockdown: LockdownClient, loop=None):
        if loop is None:
            try:
                # When deprecating python3.6, replace with get_running_loop.
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()

        nest_asyncio.apply()
        self.loop = loop
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.service = None  # type: ServiceConnection
        self.connection_id = str(uuid.uuid4()).upper()
        self.state = None
        self.connected_application = {}
        self.application_pages = {}
        self.wir_message_results = {}
        self.wir_events = []
        self.receive_handlers = {
            '_rpc_reportCurrentState:': self._handle_report_current_state,
            '_rpc_reportConnectedApplicationList:': self._handle_report_connected_application_list,
            '_rpc_reportConnectedDriverList:': self._handle_report_connected_driver_list,
            '_rpc_applicationSentListing:': self._handle_application_sent_listing,
            '_rpc_applicationUpdated:': self._handle_application_updated,
            '_rpc_applicationConnected:': self._handle_application_connected,
            '_rpc_applicationSentData:': self._handle_application_sent_data,
            '_rpc_applicationDisconnected:': self._handle_application_disconnected,
        }
        self._recv_task = None  # type: asyncio.Task | None

    def connect(self, timeout=None):
        self.service = self.await_(self.lockdown.aio_start_service(self.SERVICE_NAME))
        self.await_(self._report_identifier())
        try:
            self._handle_recv(self.await_(asyncio.wait_for(self._recv_message(), timeout)))
        except asyncio.TimeoutError as e:
            raise WebInspectorNotEnabled from e
        self._recv_task = self.loop.create_task(self._receiving_task())

    def close(self):
        self._recv_task.cancel()
        try:
            self.await_(self._recv_task)
        except asyncio.CancelledError:
            pass
        self.await_(self.service.aio_close())

    async def _recv_message(self):
        while True:
            try:
                return await self.service.aio_recv_plist()
            except asyncio.IncompleteReadError:
                await asyncio.sleep(0)

    async def _receiving_task(self):
        while True:
            self._handle_recv(await self._recv_message())

    def automation_session(self, app):
        if self.state == 'WIRAutomationAvailabilityNotAvailable':
            raise RemoteAutomationNotEnabled()
        session_id = str(uuid.uuid4()).upper()
        self.await_(self._forward_automation_session_request(session_id, app.id_))
        self.await_(self._forward_get_listing(app.id_))
        page = self.await_(self._wait_for_page(session_id))
        self.await_(self._forward_socket_setup(session_id, app.id_, page.id_))
        self.await_(self._forward_get_listing(app.id_))
        while not page.automation_connection_id:
            self.await_(asyncio.sleep(0))
        return AutomationSession(SessionProtocol(self, session_id, app, page))

    def get_open_pages(self):
        apps = {}
        self.await_(asyncio.gather(*[self._forward_get_listing(app) for app in self.connected_application]))
        for app in self.connected_application:
            if self.application_pages.get(app, False):
                apps[self.connected_application[app].name] = self.application_pages[app].values()
        return apps

    def open_app(self, bundle) -> Application:
        self.await_(self._request_application_launch(bundle))
        self.get_open_pages()
        return self.await_(self._wait_for_application(bundle))

    async def send_socket_data(self, session_id, app_id, page_id, data):
        await self._forward_socket_data(session_id, app_id, page_id, data)

    async def start_cdp(self, session_id, app_id, page_id):
        await self._forward_socket_setup(session_id, app_id, page_id, pause=False)

    def find_page_id(self, page_id):
        for app_id in self.application_pages:
            for page in self.application_pages[app_id]:
                if page == page_id:
                    return self.connected_application[app_id], self.application_pages[app_id][page_id]

    def flush_input(self, duration=0):
        return self.await_(asyncio.sleep(duration))

    def await_(self, awaitable):
        return self.loop.run_until_complete(asyncio.ensure_future(awaitable, loop=self.loop))

    def _handle_recv(self, plist):
        self.receive_handlers[plist['__selector']](plist['__argument'])

    def _handle_report_current_state(self, arg):
        self.state = arg['WIRAutomationAvailabilityKey']

    def _handle_report_connected_application_list(self, arg):
        self.connected_application = {}
        for key, application in arg['WIRApplicationDictionaryKey'].items():
            self.connected_application[key] = Application.from_application_dictionary(application)

    def _handle_report_connected_driver_list(self, arg):
        pass

    def _handle_application_sent_listing(self, arg):
        if arg['WIRApplicationIdentifierKey'] in self.application_pages:
            for id_, page in arg['WIRListingKey'].items():
                if id_ in self.application_pages[arg['WIRApplicationIdentifierKey']]:
                    self.application_pages[arg['WIRApplicationIdentifierKey']][id_].update(page)
                else:
                    self.application_pages[arg['WIRApplicationIdentifierKey']][id_] = Page.from_page_dictionary(page)
        else:
            pages = {}
            for id_, page in arg['WIRListingKey'].items():
                pages[id_] = Page.from_page_dictionary(page)
            self.application_pages[arg['WIRApplicationIdentifierKey']] = pages

    def _handle_application_updated(self, arg):
        app = Application.from_application_dictionary(arg)
        self.connected_application[app.id_] = app

    def _handle_application_connected(self, arg):
        app = Application.from_application_dictionary(arg)
        self.connected_application[app.id_] = app

    def _handle_application_sent_data(self, arg):
        response = json.loads(arg['WIRMessageDataKey'])
        if 'id' in response:
            self.wir_message_results[response['id']] = response
        else:
            self.wir_events.append(response)

    def _handle_application_disconnected(self, arg):
        self.connected_application.pop(arg['WIRApplicationIdentifierKey'], None)
        self.application_pages.pop(arg['WIRApplicationIdentifierKey'], None)

    async def _report_identifier(self):
        await self._send_message('_rpc_reportIdentifier:')

    async def _forward_get_listing(self, app_id):
        self.logger.debug(f'Listing app with id {app_id}')
        await self._send_message('_rpc_forwardGetListing:', {'WIRApplicationIdentifierKey': app_id})

    async def _request_application_launch(self, bundle):
        await self._send_message('_rpc_requestApplicationLaunch:', {'WIRApplicationBundleIdentifierKey': bundle})

    async def _forward_automation_session_request(self, session_id, app_id):
        await self._send_message('_rpc_forwardAutomationSessionRequest:', {
            'WIRApplicationIdentifierKey': app_id,
            'WIRSessionCapabilitiesKey': {
                'org.webkit.webdriver.webrtc.allow-insecure-media-capture': True,
                'org.webkit.webdriver.webrtc.suppress-ice-candidate-filtering': False,
            },
            'WIRSessionIdentifierKey': session_id
        })

    async def _forward_socket_setup(self, session_id, app_id, page_id, pause=True):
        message = {
            'WIRApplicationIdentifierKey': app_id,
            'WIRPageIdentifierKey': page_id,
            'WIRSenderKey': session_id
        }
        if not pause:
            message['WIRAutomaticallyPause'] = False
        await self._send_message('_rpc_forwardSocketSetup:', message)

    async def _forward_socket_data(self, session_id, app_id, page_id, data):
        await self._send_message('_rpc_forwardSocketData:', {
            'WIRApplicationIdentifierKey': app_id,
            'WIRPageIdentifierKey': page_id,
            'WIRSessionIdentifierKey': session_id,
            'WIRSocketDataKey': json.dumps(data).encode(),
        })

    async def _forward_indicate_web_view(self, app_id, page_id, enable: bool):
        await self._send_message('_rpc_forwardIndicateWebView'), {
            'WIRApplicationIdentifierKey': app_id,
            'WIRPageIdentifierKey': page_id,
            'WIRIndicateEnabledKey': enable,
        }

    async def _send_message(self, selector, args=None):
        if args is None:
            args = {}
        args['WIRConnectionIdentifierKey'] = self.connection_id
        await self.service.aio_send_plist({'__selector': selector, '__argument': args})

    def _page_by_automation_session(self, session_id):
        for app_id in self.application_pages:
            for page in self.application_pages[app_id]:
                if page.type_ == WirTypes.AUTOMATION and page.automation_session_id == session_id:
                    return page

    async def _wait_for_page(self, session_id):
        while True:
            for app in self.application_pages.values():
                for page in app.values():
                    if page.type_ == WirTypes.AUTOMATION and page.automation_session_id == session_id:
                        return page
            await asyncio.sleep(0)

    async def _wait_for_application(self, bundle='', app_id=''):
        while True:
            for app in self.connected_application.values():
                if bundle and app.bundle == bundle:
                    return app
                if app_id and app.id_ == app_id:
                    return app
            await asyncio.sleep(0)



from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class ScreenshotService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.screenshotr'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME, is_developer_service=True)

        dl_message_version_exchange = self.service.recv_plist()
        version_major = dl_message_version_exchange[1]
        dl_message_device_ready = self.service.send_recv_plist(
            ['DLMessageVersionExchange', 'DLVersionsOk', version_major])
        if dl_message_device_ready[0] != 'DLMessageDeviceReady':
            raise PyMobileDevice3Exception('Screenshotr didn\'t return ready state')

    def take_screenshot(self) -> bytes:
        self.service.send_plist(['DLMessageProcessMessage', {'MessageType': 'ScreenShotRequest'}])
        response = self.service.recv_plist()

        assert len(response) == 2
        assert response[0] == 'DLMessageProcessMessage'

        if response[1].get('MessageType') == 'ScreenShotReply':
            return response[1]['ScreenShotData']

        raise PyMobileDevice3Exception(f'invalid response: {response}')



from pymobiledevice3.services.remote_server import MessageAux


class ApplicationListing:
    IDENTIFIER = 'com.apple.instruments.server.services.device.applictionListing'

    def __init__(self, dvt):
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def applist(self) -> list:
        """
        Get the applications list from the device.
        :return: List of applications and their attributes.
        """
        self._channel.installedApplicationsMatching_registerUpdateToken_(
            MessageAux().append_obj({}).append_obj(''))
        result = self._channel.receive_plist()
        assert isinstance(result, list)
        return result



import datetime
import ctypes
import shutil
import os
import struct
from pathlib import Path

from pymobiledevice3.exceptions import PyMobileDevice3Exception

SIZE_FORMAT = '>I'
CODE_FORMAT = '>B'
CODE_FILE_DATA = 0xc
CODE_ERROR_LOCAL = 0x6
CODE_SUCCESS = 0
FILE_TRANSFER_TERMINATOR = b'\x00\x00\x00\x00'
BULK_OPERATION_ERROR = -13
APPLE_EPOCH = 978307200
ERRNO_TO_DEVICE_ERROR = {
    2: -6,
    17: -7,
    20: -8,
    21: -9,
    62: -10,
    5: -11,
    28: -15,
}


class DeviceLink:
    def __init__(self, service, root_path: Path):
        self.service = service
        self.root_path = root_path
        self._dl_handlers = {
            'DLMessageCreateDirectory': self.create_directory,
            'DLMessageUploadFiles': self.upload_files,
            'DLMessageGetFreeDiskSpace': self.get_free_disk_space,
            'DLMessageMoveItems': self.move_items,
            'DLMessageRemoveItems': self.remove_items,
            'DLMessageDownloadFiles': self.download_files,
            'DLContentsOfDirectory': self.contents_of_directory,
            'DLMessageCopyItem': self.copy_item,
        }

    def dl_loop(self, progress_callback=lambda x: None):
        while True:
            message = self.receive_message()
            command = message[0]

            if command in ('DLMessageDownloadFiles', 'DLMessageMoveFiles', 'DLMessageMoveItems', 'DLMessageRemoveFiles',
                           'DLMessageRemoveItems'):
                progress_callback(message[3])
            elif command == 'DLMessageUploadFiles':
                progress_callback(message[2])

            if command == 'DLMessageProcessMessage':
                if not message[1]['ErrorCode']:
                    return message[1].get('Content')
                else:
                    raise PyMobileDevice3Exception(f'Device link error: {message[1]}')
            self._dl_handlers[command](message)

    def version_exchange(self):
        dl_message_version_exchange = self.receive_message()
        version_major = dl_message_version_exchange[1]
        self.service.send_plist(['DLMessageVersionExchange', 'DLVersionsOk', version_major])
        dl_message_device_ready = self.receive_message()
        if dl_message_device_ready[0] != 'DLMessageDeviceReady':
            raise PyMobileDevice3Exception('Device link didn\'t return ready state')

    def send_process_message(self, message):
        self.service.send_plist(['DLMessageProcessMessage', message])

    def download_files(self, message):
        status = {}
        for file in message[1]:
            self.service.sendall(struct.pack(SIZE_FORMAT, len(file)))
            self.service.sendall(file.encode())
            try:
                data = (self.root_path / file).read_bytes()
            except IOError as e:
                status[file] = {
                    'DLFileErrorString': e.strerror,
                    'DLFileErrorCode': ctypes.c_uint64(ERRNO_TO_DEVICE_ERROR[e.errno]).value
                }
                self.service.sendall(struct.pack(SIZE_FORMAT, len(e.strerror) + struct.calcsize(CODE_FORMAT)))
                self.service.sendall(struct.pack(CODE_FORMAT, CODE_ERROR_LOCAL) + e.strerror.encode())
            else:
                self.service.sendall(struct.pack(SIZE_FORMAT, len(data) + struct.calcsize(CODE_FORMAT)))
                self.service.sendall(struct.pack(CODE_FORMAT, CODE_FILE_DATA) + data)
                buffer = struct.pack(SIZE_FORMAT, struct.calcsize(CODE_FORMAT)) + struct.pack(CODE_FORMAT, CODE_SUCCESS)
                self.service.sendall(buffer)
        self.service.sendall(FILE_TRANSFER_TERMINATOR)
        if status:
            self.status_response(BULK_OPERATION_ERROR, 'Multi status', status)
        else:
            self.status_response(0)

    def contents_of_directory(self, message):
        data = {}
        path = self.root_path / message[1]
        for file in path.iterdir():
            ftype = 'DLFileTypeUnknown'
            if file.is_dir():
                ftype = 'DLFileTypeDirectory'
            if file.is_file():
                ftype = 'DLFileTypeRegular'
            data[file.name] = {
                'DLFileType': ftype,
                'DLFileSize': file.stat().st_size,
                'DLFileModificationDate': datetime.datetime.utcfromtimestamp(file.stat().st_mtime - APPLE_EPOCH)
            }
        self.status_response(0, status_dict=data)

    def upload_files(self, message):
        real_size = 0
        while True:
            device_name = self._prefixed_recv()
            if not device_name:
                break
            file_name = self._prefixed_recv()
            size, = struct.unpack(SIZE_FORMAT, self.service.recvall(struct.calcsize(SIZE_FORMAT)))
            code, = struct.unpack(CODE_FORMAT, self.service.recvall(struct.calcsize(CODE_FORMAT)))
            size -= struct.calcsize(CODE_FORMAT)
            data = b''
            while size and code == CODE_FILE_DATA:
                real_size += size
                data += self.service.recvall(size)
                size, = struct.unpack(SIZE_FORMAT, self.service.recvall(struct.calcsize(SIZE_FORMAT)))
                code, = struct.unpack(CODE_FORMAT, self.service.recvall(struct.calcsize(CODE_FORMAT)))
                size -= struct.calcsize(CODE_FORMAT)
            assert code == CODE_SUCCESS
            (self.root_path / file_name).write_bytes(data)
        self.status_response(0)

    def get_free_disk_space(self, message):
        vfs = os.statvfs(self.root_path)
        freespace = vfs.f_bavail * vfs.f_bsize
        self.status_response(0, status_dict=freespace)

    def move_items(self, message):
        for src, dst in message[1].items():
            dest = self.root_path / dst
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.move(self.root_path / src, dest)
        self.status_response(0)

    def copy_item(self, message):
        src = self.root_path / message[1]
        dest = self.root_path / message[2]
        dest.parent.mkdir(parents=True, exist_ok=True)
        if src.is_dir():
            shutil.copytree(src, dest)
        else:
            shutil.copy(src, dest)
        self.status_response(0)

    def remove_items(self, message):
        for path in message[1]:
            rm_path = self.root_path / path
            if rm_path.is_dir():
                shutil.rmtree(rm_path)
            else:
                rm_path.unlink(missing_ok=True)
        self.status_response(0)

    def create_directory(self, message):
        path = message[1]
        (self.root_path / path).mkdir(parents=True, exist_ok=True)
        self.status_response(0)

    def status_response(self, status_code, status_str='', status_dict=None):
        self.service.send_plist([
            'DLMessageStatusResponse', ctypes.c_uint64(status_code).value,
            status_str if status_str else '___EmptyParameterString___',
            status_dict if status_dict is not None else {},
        ])

    def receive_message(self):
        return self.service.recv_plist()

    def disconnect(self):
        self.service.send_plist(['DLMessageDisconnect', '___EmptyParameterString___'])

    def _prefixed_recv(self):
        size, = struct.unpack(SIZE_FORMAT, self.service.recvall(struct.calcsize(SIZE_FORMAT)))
        return self.service.recvall(size).decode()
